@@ -11,14 +11,15 @@ use crate::debuggers::{extract_cdb_version, extract_gdb_version};
 pub(crate) use crate::directives::auxiliary::AuxProps;
 use crate::directives::auxiliary::parse_and_update_aux;
 use crate::directives::directive_names::{
-    KNOWN_DIRECTIVE_NAMES, KNOWN_HTMLDOCCK_DIRECTIVE_NAMES, KNOWN_JSONDOCCK_DIRECTIVE_NAMES,
+    KNOWN_DIRECTIVE_NAMES_SET, KNOWN_HTMLDOCCK_DIRECTIVE_NAMES, KNOWN_JSONDOCCK_DIRECTIVE_NAMES,
 };
 pub(crate) use crate::directives::file::FileDirectives;
+use crate::directives::handlers::DIRECTIVE_HANDLERS_MAP;
 use crate::directives::line::{DirectiveLine, line_directive};
 use crate::directives::needs::CachedNeedsConditions;
 use crate::edition::{Edition, parse_edition};
 use crate::errors::ErrorKind;
-use crate::executor::{CollectedTestDesc, ShouldPanic};
+use crate::executor::{CollectedTestDesc, ShouldFail};
 use crate::util::static_regex;
 use crate::{fatal, help};
 
@@ -26,18 +27,25 @@ mod auxiliary;
 mod cfg;
 mod directive_names;
 mod file;
+mod handlers;
 mod line;
 mod needs;
 #[cfg(test)]
 mod tests;
 
 pub struct DirectivesCache {
+    /// "Conditions" used by `ignore-*` and `only-*` directives, prepared in
+    /// advance so that they don't have to be evaluated repeatedly.
+    cfg_conditions: cfg::PreparedConditions,
     needs: CachedNeedsConditions,
 }
 
 impl DirectivesCache {
     pub fn load(config: &Config) -> Self {
-        Self { needs: CachedNeedsConditions::load(config) }
+        Self {
+            cfg_conditions: cfg::prepare_conditions(config),
+            needs: CachedNeedsConditions::load(config),
+        }
     }
 }
 
@@ -53,13 +61,10 @@ impl EarlyProps {
         config: &Config,
         file_directives: &FileDirectives<'_>,
     ) -> Self {
-        let testfile = file_directives.path;
         let mut props = EarlyProps::default();
-        let mut poisoned = false;
 
         iter_directives(
             config.mode,
-            &mut poisoned,
             file_directives,
             // (dummy comment to force args into vertical layout)
             &mut |ln: &DirectiveLine<'_>| {
@@ -67,11 +72,6 @@ impl EarlyProps {
             },
         );
 
-        if poisoned {
-            eprintln!("errors encountered during EarlyProps parsing: {}", testfile);
-            panic!("errors encountered during EarlyProps parsing");
-        }
-
         props
     }
 }
@@ -200,9 +200,9 @@ pub(crate) struct TestProps {
     pub no_auto_check_cfg: bool,
     /// Build and use `minicore` as `core` stub for `no_core` tests in cross-compilation scenarios
     /// that don't otherwise want/need `-Z build-std`.
-    pub add_core_stubs: bool,
+    pub add_minicore: bool,
     /// Add these flags to the build of `minicore`.
-    pub core_stubs_compile_flags: Vec<String>,
+    pub minicore_compile_flags: Vec<String>,
     /// Whether line annotatins are required for the given error kind.
     pub dont_require_annotations: HashSet<ErrorKind>,
     /// Whether pretty printers should be disabled in gdb.
@@ -254,8 +254,8 @@ mod directives {
     pub const LLVM_COV_FLAGS: &'static str = "llvm-cov-flags";
     pub const FILECHECK_FLAGS: &'static str = "filecheck-flags";
     pub const NO_AUTO_CHECK_CFG: &'static str = "no-auto-check-cfg";
-    pub const ADD_CORE_STUBS: &'static str = "add-core-stubs";
-    pub const CORE_STUBS_COMPILE_FLAGS: &'static str = "core-stubs-compile-flags";
+    pub const ADD_MINICORE: &'static str = "add-minicore";
+    pub const MINICORE_COMPILE_FLAGS: &'static str = "minicore-compile-flags";
     pub const DISABLE_GDB_PRETTY_PRINTERS: &'static str = "disable-gdb-pretty-printers";
     pub const COMPARE_OUTPUT_BY_LINES: &'static str = "compare-output-by-lines";
 }
@@ -311,8 +311,8 @@ impl TestProps {
             llvm_cov_flags: vec![],
             filecheck_flags: vec![],
             no_auto_check_cfg: false,
-            add_core_stubs: false,
-            core_stubs_compile_flags: vec![],
+            add_minicore: false,
+            minicore_compile_flags: vec![],
             dont_require_annotations: Default::default(),
             disable_gdb_pretty_printers: false,
             compare_output_by_lines: false,
@@ -358,287 +358,20 @@ impl TestProps {
             let file_contents = fs::read_to_string(testfile).unwrap();
             let file_directives = FileDirectives::from_file_contents(testfile, &file_contents);
 
-            let mut poisoned = false;
-
             iter_directives(
                 config.mode,
-                &mut poisoned,
                 &file_directives,
+                // (dummy comment to force args into vertical layout)
                 &mut |ln: &DirectiveLine<'_>| {
                     if !ln.applies_to_test_revision(test_revision) {
                         return;
                     }
 
-                    use directives::*;
-
-                    config.push_name_value_directive(
-                        ln,
-                        ERROR_PATTERN,
-                        &mut self.error_patterns,
-                        |r| r,
-                    );
-                    config.push_name_value_directive(
-                        ln,
-                        REGEX_ERROR_PATTERN,
-                        &mut self.regex_error_patterns,
-                        |r| r,
-                    );
-
-                    config.push_name_value_directive(ln, DOC_FLAGS, &mut self.doc_flags, |r| r);
-
-                    fn split_flags(flags: &str) -> Vec<String> {
-                        // Individual flags can be single-quoted to preserve spaces; see
-                        // <https://github.com/rust-lang/rust/pull/115948/commits/957c5db6>.
-                        flags
-                            .split('\'')
-                            .enumerate()
-                            .flat_map(|(i, f)| {
-                                if i % 2 == 1 { vec![f] } else { f.split_whitespace().collect() }
-                            })
-                            .map(move |s| s.to_owned())
-                            .collect::<Vec<_>>()
+                    if let Some(handler) = DIRECTIVE_HANDLERS_MAP.get(ln.name) {
+                        handler.handle(config, ln, self);
                     }
-
-                    if let Some(flags) = config.parse_name_value_directive(ln, COMPILE_FLAGS) {
-                        let flags = split_flags(&flags);
-                        for (i, flag) in flags.iter().enumerate() {
-                            if flag == "--edition" || flag.starts_with("--edition=") {
-                                panic!("you must use `//@ edition` to configure the edition");
-                            }
-                            if (flag == "-C"
-                                && flags.get(i + 1).is_some_and(|v| v.starts_with("incremental=")))
-                                || flag.starts_with("-Cincremental=")
-                            {
-                                panic!(
-                                    "you must use `//@ incremental` to enable incremental compilation"
-                                );
-                            }
-                        }
-                        self.compile_flags.extend(flags);
-                    }
-
-                    if let Some(range) = parse_edition_range(config, ln) {
-                        self.edition = Some(range.edition_to_test(config.edition));
-                    }
-
-                    config.parse_and_update_revisions(ln, &mut self.revisions);
-
-                    if let Some(flags) = config.parse_name_value_directive(ln, RUN_FLAGS) {
-                        self.run_flags.extend(split_flags(&flags));
-                    }
-
-                    if self.pp_exact.is_none() {
-                        self.pp_exact = config.parse_pp_exact(ln);
-                    }
-
-                    config.set_name_directive(ln, SHOULD_ICE, &mut self.should_ice);
-                    config.set_name_directive(ln, BUILD_AUX_DOCS, &mut self.build_aux_docs);
-                    config.set_name_directive(ln, UNIQUE_DOC_OUT_DIR, &mut self.unique_doc_out_dir);
-
-                    config.set_name_directive(ln, FORCE_HOST, &mut self.force_host);
-                    config.set_name_directive(ln, CHECK_STDOUT, &mut self.check_stdout);
-                    config.set_name_directive(ln, CHECK_RUN_RESULTS, &mut self.check_run_results);
-                    config.set_name_directive(
-                        ln,
-                        DONT_CHECK_COMPILER_STDOUT,
-                        &mut self.dont_check_compiler_stdout,
-                    );
-                    config.set_name_directive(
-                        ln,
-                        DONT_CHECK_COMPILER_STDERR,
-                        &mut self.dont_check_compiler_stderr,
-                    );
-                    config.set_name_directive(ln, NO_PREFER_DYNAMIC, &mut self.no_prefer_dynamic);
-
-                    if let Some(m) = config.parse_name_value_directive(ln, PRETTY_MODE) {
-                        self.pretty_mode = m;
-                    }
-
-                    config.set_name_directive(
-                        ln,
-                        PRETTY_COMPARE_ONLY,
-                        &mut self.pretty_compare_only,
-                    );
-
-                    // Call a helper method to deal with aux-related directives.
-                    parse_and_update_aux(config, ln, &mut self.aux);
-
-                    config.push_name_value_directive(
-                        ln,
-                        EXEC_ENV,
-                        &mut self.exec_env,
-                        Config::parse_env,
-                    );
-                    config.push_name_value_directive(
-                        ln,
-                        UNSET_EXEC_ENV,
-                        &mut self.unset_exec_env,
-                        |r| r.trim().to_owned(),
-                    );
-                    config.push_name_value_directive(
-                        ln,
-                        RUSTC_ENV,
-                        &mut self.rustc_env,
-                        Config::parse_env,
-                    );
-                    config.push_name_value_directive(
-                        ln,
-                        UNSET_RUSTC_ENV,
-                        &mut self.unset_rustc_env,
-                        |r| r.trim().to_owned(),
-                    );
-                    config.push_name_value_directive(
-                        ln,
-                        FORBID_OUTPUT,
-                        &mut self.forbid_output,
-                        |r| r,
-                    );
-                    config.set_name_directive(
-                        ln,
-                        CHECK_TEST_LINE_NUMBERS_MATCH,
-                        &mut self.check_test_line_numbers_match,
-                    );
-
-                    self.update_pass_mode(ln, config);
-                    self.update_fail_mode(ln, config);
-
-                    config.set_name_directive(ln, IGNORE_PASS, &mut self.ignore_pass);
-
-                    if let Some(NormalizeRule { kind, regex, replacement }) =
-                        config.parse_custom_normalization(ln)
-                    {
-                        let rule_tuple = (regex, replacement);
-                        match kind {
-                            NormalizeKind::Stdout => self.normalize_stdout.push(rule_tuple),
-                            NormalizeKind::Stderr => self.normalize_stderr.push(rule_tuple),
-                            NormalizeKind::Stderr32bit => {
-                                if config.target_cfg().pointer_width == 32 {
-                                    self.normalize_stderr.push(rule_tuple);
-                                }
-                            }
-                            NormalizeKind::Stderr64bit => {
-                                if config.target_cfg().pointer_width == 64 {
-                                    self.normalize_stderr.push(rule_tuple);
-                                }
-                            }
-                        }
-                    }
-
-                    if let Some(code) = config
-                        .parse_name_value_directive(ln, FAILURE_STATUS)
-                        .and_then(|code| code.trim().parse::<i32>().ok())
-                    {
-                        self.failure_status = Some(code);
-                    }
-
-                    config.set_name_directive(
-                        ln,
-                        DONT_CHECK_FAILURE_STATUS,
-                        &mut self.dont_check_failure_status,
-                    );
-
-                    config.set_name_directive(ln, RUN_RUSTFIX, &mut self.run_rustfix);
-                    config.set_name_directive(
-                        ln,
-                        RUSTFIX_ONLY_MACHINE_APPLICABLE,
-                        &mut self.rustfix_only_machine_applicable,
-                    );
-                    config.set_name_value_directive(
-                        ln,
-                        ASSEMBLY_OUTPUT,
-                        &mut self.assembly_output,
-                        |r| r.trim().to_string(),
-                    );
-                    config.set_name_directive(
-                        ln,
-                        STDERR_PER_BITWIDTH,
-                        &mut self.stderr_per_bitwidth,
-                    );
-                    config.set_name_directive(ln, INCREMENTAL, &mut self.incremental);
-
-                    // Unlike the other `name_value_directive`s this needs to be handled manually,
-                    // because it sets a `bool` flag.
-                    if let Some(known_bug) = config.parse_name_value_directive(ln, KNOWN_BUG) {
-                        let known_bug = known_bug.trim();
-                        if known_bug == "unknown"
-                            || known_bug.split(',').all(|issue_ref| {
-                                issue_ref
-                                    .trim()
-                                    .split_once('#')
-                                    .filter(|(_, number)| {
-                                        number.chars().all(|digit| digit.is_numeric())
-                                    })
-                                    .is_some()
-                            })
-                        {
-                            self.known_bug = true;
-                        } else {
-                            panic!(
-                                "Invalid known-bug value: {known_bug}\nIt requires comma-separated issue references (`#000` or `chalk#000`) or `known-bug: unknown`."
-                            );
-                        }
-                    } else if config.parse_name_directive(ln, KNOWN_BUG) {
-                        panic!(
-                            "Invalid known-bug attribute, requires comma-separated issue references (`#000` or `chalk#000`) or `known-bug: unknown`."
-                        );
-                    }
-
-                    config.set_name_value_directive(
-                        ln,
-                        TEST_MIR_PASS,
-                        &mut self.mir_unit_test,
-                        |s| s.trim().to_string(),
-                    );
-                    config.set_name_directive(ln, REMAP_SRC_BASE, &mut self.remap_src_base);
-
-                    if let Some(flags) = config.parse_name_value_directive(ln, LLVM_COV_FLAGS) {
-                        self.llvm_cov_flags.extend(split_flags(&flags));
-                    }
-
-                    if let Some(flags) = config.parse_name_value_directive(ln, FILECHECK_FLAGS) {
-                        self.filecheck_flags.extend(split_flags(&flags));
-                    }
-
-                    config.set_name_directive(ln, NO_AUTO_CHECK_CFG, &mut self.no_auto_check_cfg);
-
-                    self.update_add_core_stubs(ln, config);
-
-                    if let Some(flags) =
-                        config.parse_name_value_directive(ln, CORE_STUBS_COMPILE_FLAGS)
-                    {
-                        let flags = split_flags(&flags);
-                        for flag in &flags {
-                            if flag == "--edition" || flag.starts_with("--edition=") {
-                                panic!("you must use `//@ edition` to configure the edition");
-                            }
-                        }
-                        self.core_stubs_compile_flags.extend(flags);
-                    }
-
-                    if let Some(err_kind) =
-                        config.parse_name_value_directive(ln, DONT_REQUIRE_ANNOTATIONS)
-                    {
-                        self.dont_require_annotations
-                            .insert(ErrorKind::expect_from_user_str(err_kind.trim()));
-                    }
-
-                    config.set_name_directive(
-                        ln,
-                        DISABLE_GDB_PRETTY_PRINTERS,
-                        &mut self.disable_gdb_pretty_printers,
-                    );
-                    config.set_name_directive(
-                        ln,
-                        COMPARE_OUTPUT_BY_LINES,
-                        &mut self.compare_output_by_lines,
-                    );
                 },
             );
-
-            if poisoned {
-                eprintln!("errors encountered during TestProps parsing: {}", testfile);
-                panic!("errors encountered during TestProps parsing");
-            }
         }
 
         if self.should_ice {
@@ -753,12 +486,12 @@ impl TestProps {
         self.pass_mode
     }
 
-    fn update_add_core_stubs(&mut self, ln: &DirectiveLine<'_>, config: &Config) {
-        let add_core_stubs = config.parse_name_directive(ln, directives::ADD_CORE_STUBS);
-        if add_core_stubs {
+    fn update_add_minicore(&mut self, ln: &DirectiveLine<'_>, config: &Config) {
+        let add_minicore = config.parse_name_directive(ln, directives::ADD_MINICORE);
+        if add_minicore {
             if !matches!(config.mode, TestMode::Ui | TestMode::Codegen | TestMode::Assembly) {
                 panic!(
-                    "`add-core-stubs` is currently only supported for ui, codegen and assembly test modes"
+                    "`add-minicore` is currently only supported for ui, codegen and assembly test modes"
                 );
             }
 
@@ -767,14 +500,42 @@ impl TestProps {
             if self.local_pass_mode().is_some_and(|pm| pm == PassMode::Run) {
                 // `minicore` can only be used with non-run modes, because it's `core` prelude stubs
                 // and can't run.
-                panic!("`add-core-stubs` cannot be used to run the test binary");
+                panic!("`add-minicore` cannot be used to run the test binary");
             }
 
-            self.add_core_stubs = add_core_stubs;
+            self.add_minicore = add_minicore;
         }
     }
 }
 
+pub(crate) fn do_early_directives_check(
+    mode: TestMode,
+    file_directives: &FileDirectives<'_>,
+) -> Result<(), String> {
+    let testfile = file_directives.path;
+
+    for directive_line @ DirectiveLine { line_number, .. } in &file_directives.lines {
+        let CheckDirectiveResult { is_known_directive, trailing_directive } =
+            check_directive(directive_line, mode);
+
+        if !is_known_directive {
+            return Err(format!(
+                "ERROR: unknown compiletest directive `{directive}` at {testfile}:{line_number}",
+                directive = directive_line.display(),
+            ));
+        }
+
+        if let Some(trailing_directive) = &trailing_directive {
+            return Err(format!(
+                "ERROR: detected trailing compiletest directive `{trailing_directive}` at {testfile}:{line_number}\n\
+                HELP: put the directive on its own line: `//@ {trailing_directive}`"
+            ));
+        }
+    }
+
+    Ok(())
+}
+
 pub(crate) struct CheckDirectiveResult<'ln> {
     is_known_directive: bool,
     trailing_directive: Option<&'ln str>,
@@ -786,7 +547,7 @@ fn check_directive<'a>(
 ) -> CheckDirectiveResult<'a> {
     let &DirectiveLine { name: directive_name, .. } = directive_ln;
 
-    let is_known_directive = KNOWN_DIRECTIVE_NAMES.contains(&directive_name)
+    let is_known_directive = KNOWN_DIRECTIVE_NAMES_SET.contains(&directive_name)
         || match mode {
             TestMode::Rustdoc => KNOWN_HTMLDOCCK_DIRECTIVE_NAMES.contains(&directive_name),
             TestMode::RustdocJson => KNOWN_JSONDOCCK_DIRECTIVE_NAMES.contains(&directive_name),
@@ -799,7 +560,7 @@ fn check_directive<'a>(
     let trailing_directive = directive_ln
         .remark_after_space()
         .map(|remark| remark.trim_start().split(' ').next().unwrap())
-        .filter(|token| KNOWN_DIRECTIVE_NAMES.contains(token));
+        .filter(|token| KNOWN_DIRECTIVE_NAMES_SET.contains(token));
 
     // FIXME(Zalathar): Consider emitting specialized error/help messages for
     // bogus directive names that are similar to real ones, e.g.:
@@ -811,7 +572,6 @@ fn check_directive<'a>(
 
 fn iter_directives(
     mode: TestMode,
-    poisoned: &mut bool,
     file_directives: &FileDirectives<'_>,
     it: &mut dyn FnMut(&DirectiveLine<'_>),
 ) {
@@ -837,36 +597,7 @@ fn iter_directives(
         }
     }
 
-    for directive_line @ &DirectiveLine { line_number, .. } in &file_directives.lines {
-        // Perform unknown directive check on Rust files.
-        if testfile.extension() == Some("rs") {
-            let CheckDirectiveResult { is_known_directive, trailing_directive } =
-                check_directive(directive_line, mode);
-
-            if !is_known_directive {
-                *poisoned = true;
-
-                error!(
-                    "{testfile}:{line_number}: detected unknown compiletest test directive `{}`",
-                    directive_line.display(),
-                );
-
-                return;
-            }
-
-            if let Some(trailing_directive) = &trailing_directive {
-                *poisoned = true;
-
-                error!(
-                    "{testfile}:{line_number}: detected trailing compiletest test directive `{}`",
-                    trailing_directive,
-                );
-                help!("put the trailing directive in its own line: `//@ {}`", trailing_directive);
-
-                return;
-            }
-        }
-
+    for directive_line in &file_directives.lines {
         it(directive_line);
     }
 }
@@ -1154,107 +885,6 @@ pub fn extract_llvm_version_from_binary(binary_path: &str) -> Option<Version> {
     None
 }
 
-/// For tests using the `needs-llvm-zstd` directive:
-/// - for local LLVM builds, try to find the static zstd library in the llvm-config system libs.
-/// - for `download-ci-llvm`, see if `lld` was built with zstd support.
-pub fn llvm_has_libzstd(config: &Config) -> bool {
-    // Strategy 1: works for local builds but not with `download-ci-llvm`.
-    //
-    // We check whether `llvm-config` returns the zstd library. Bootstrap's `llvm.libzstd` will only
-    // ask to statically link it when building LLVM, so we only check if the list of system libs
-    // contains a path to that static lib, and that it exists.
-    //
-    // See compiler/rustc_llvm/build.rs for more details and similar expectations.
-    fn is_zstd_in_config(llvm_bin_dir: &Utf8Path) -> Option<()> {
-        let llvm_config_path = llvm_bin_dir.join("llvm-config");
-        let output = Command::new(llvm_config_path).arg("--system-libs").output().ok()?;
-        assert!(output.status.success(), "running llvm-config --system-libs failed");
-
-        let libs = String::from_utf8(output.stdout).ok()?;
-        for lib in libs.split_whitespace() {
-            if lib.ends_with("libzstd.a") && Utf8Path::new(lib).exists() {
-                return Some(());
-            }
-        }
-
-        None
-    }
-
-    // Strategy 2: `download-ci-llvm`'s `llvm-config --system-libs` will not return any libs to
-    // use.
-    //
-    // The CI artifacts also don't contain the bootstrap config used to build them: otherwise we
-    // could have looked at the `llvm.libzstd` config.
-    //
-    // We infer whether `LLVM_ENABLE_ZSTD` was used to build LLVM as a byproduct of testing whether
-    // `lld` supports it. If not, an error will be emitted: "LLVM was not built with
-    // LLVM_ENABLE_ZSTD or did not find zstd at build time".
-    #[cfg(unix)]
-    fn is_lld_built_with_zstd(llvm_bin_dir: &Utf8Path) -> Option<()> {
-        let lld_path = llvm_bin_dir.join("lld");
-        if lld_path.exists() {
-            // We can't call `lld` as-is, it expects to be invoked by a compiler driver using a
-            // different name. Prepare a temporary symlink to do that.
-            let lld_symlink_path = llvm_bin_dir.join("ld.lld");
-            if !lld_symlink_path.exists() {
-                std::os::unix::fs::symlink(lld_path, &lld_symlink_path).ok()?;
-            }
-
-            // Run `lld` with a zstd flag. We expect this command to always error here, we don't
-            // want to link actual files and don't pass any.
-            let output = Command::new(&lld_symlink_path)
-                .arg("--compress-debug-sections=zstd")
-                .output()
-                .ok()?;
-            assert!(!output.status.success());
-
-            // Look for a specific error caused by LLVM not being built with zstd support. We could
-            // also look for the "no input files" message, indicating the zstd flag was accepted.
-            let stderr = String::from_utf8(output.stderr).ok()?;
-            let zstd_available = !stderr.contains("LLVM was not built with LLVM_ENABLE_ZSTD");
-
-            // We don't particularly need to clean the link up (so the previous commands could fail
-            // in theory but won't in practice), but we can try.
-            std::fs::remove_file(lld_symlink_path).ok()?;
-
-            if zstd_available {
-                return Some(());
-            }
-        }
-
-        None
-    }
-
-    #[cfg(not(unix))]
-    fn is_lld_built_with_zstd(_llvm_bin_dir: &Utf8Path) -> Option<()> {
-        None
-    }
-
-    if let Some(llvm_bin_dir) = &config.llvm_bin_dir {
-        // Strategy 1: for local LLVM builds.
-        if is_zstd_in_config(llvm_bin_dir).is_some() {
-            return true;
-        }
-
-        // Strategy 2: for LLVM artifacts built on CI via `download-ci-llvm`.
-        //
-        // It doesn't work for cases where the artifacts don't contain the linker, but it's
-        // best-effort: CI has `llvm.libzstd` and `lld` enabled on the x64 linux artifacts, so it
-        // will at least work there.
-        //
-        // If this can be improved and expanded to less common cases in the future, it should.
-        if config.target == "x86_64-unknown-linux-gnu"
-            && config.host == config.target
-            && is_lld_built_with_zstd(llvm_bin_dir).is_some()
-        {
-            return true;
-        }
-    }
-
-    // Otherwise, all hope is lost.
-    false
-}
-
 /// Takes a directive of the form `"<version1> [- <version2>]"`, returns the numeric representation
 /// of `<version1>` and `<version2>` as tuple: `(<version1>, <version2>)`.
 ///
@@ -1304,12 +934,9 @@ pub(crate) fn make_test_description(
     let mut ignore_message = None;
     let mut should_fail = false;
 
-    let mut local_poisoned = false;
-
     // Scan through the test file to handle `ignore-*`, `only-*`, and `needs-*` directives.
     iter_directives(
         config.mode,
-        &mut local_poisoned,
         file_directives,
         &mut |ln @ &DirectiveLine { line_number, .. }| {
             if !ln.applies_to_test_revision(test_revision) {
@@ -1336,8 +963,8 @@ pub(crate) fn make_test_description(
                 };
             }
 
-            decision!(cfg::handle_ignore(config, ln));
-            decision!(cfg::handle_only(config, ln));
+            decision!(cfg::handle_ignore(&cache.cfg_conditions, ln));
+            decision!(cfg::handle_only(&cache.cfg_conditions, ln));
             decision!(needs::handle_needs(&cache.needs, config, ln));
             decision!(ignore_llvm(config, ln));
             decision!(ignore_backends(config, ln));
@@ -1358,18 +985,13 @@ pub(crate) fn make_test_description(
         },
     );
 
-    if local_poisoned {
-        eprintln!("errors encountered when trying to make test description: {}", path);
-        panic!("errors encountered when trying to make test description");
-    }
-
     // The `should-fail` annotation doesn't apply to pretty tests,
     // since we run the pretty printer across all tests by default.
     // If desired, we could add a `should-fail-pretty` annotation.
-    let should_panic = match config.mode {
-        TestMode::Pretty => ShouldPanic::No,
-        _ if should_fail => ShouldPanic::Yes,
-        _ => ShouldPanic::No,
+    let should_fail = if should_fail && config.mode != TestMode::Pretty {
+        ShouldFail::Yes
+    } else {
+        ShouldFail::No
     };
 
     CollectedTestDesc {
@@ -1377,7 +999,7 @@ pub(crate) fn make_test_description(
         filterable_path: filterable_path.to_owned(),
         ignore,
         ignore_message,
-        should_panic,
+        should_fail,
     }
 }
 
@@ -1493,7 +1115,7 @@ fn ignore_backends(config: &Config, line: &DirectiveLine<'_>) -> IgnoreDecision
                 }
             }
         }) {
-            if config.default_codegen_backend == backend {
+            if !config.bypass_ignore_backends && config.default_codegen_backend == backend {
                 return IgnoreDecision::Ignore {
                     reason: format!("{} backend is marked as ignore", backend.as_str()),
                 };
@@ -1716,3 +1338,16 @@ impl EditionRange {
         }
     }
 }
+
+fn split_flags(flags: &str) -> Vec<String> {
+    // Individual flags can be single-quoted to preserve spaces; see
+    // <https://github.com/rust-lang/rust/pull/115948/commits/957c5db6>.
+    // FIXME(#147955): Replace this ad-hoc quoting with an escape/quote system that
+    // is closer to what actual shells do, so that it's more flexible and familiar.
+    flags
+        .split('\'')
+        .enumerate()
+        .flat_map(|(i, f)| if i % 2 == 1 { vec![f] } else { f.split_whitespace().collect() })
+        .map(move |s| s.to_owned())
+        .collect::<Vec<_>>()
+}
